#include "midgard_nir.h"

#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 11
 * transforms:
 *    ('ineg', 'a') => ('isub', 0, 'a')
 *    ('b32csel', 'a', 'b@32', 0) => ('iand', 'a', 'b')
 *    ('b32csel', 'a', 0, 'b@32') => ('iand', ('inot', 'a'), 'b')
 *    ('i2i8', 'a@32') => ('i2i8', ('i2i16', 'a'))
 *    ('u2u8', 'a@32') => ('u2u8', ('u2u16', 'a'))
 *    ('i2i32', 'a@8') => ('i2i32', ('i2i16', 'a'))
 *    ('u2u32', 'a@8') => ('u2u32', ('u2u16', 'a'))
 *    ('f2i32', 'a@16') => ('f2i32', ('f2f32', 'a'))
 *    ('f2u32', 'a@16') => ('f2u32', ('f2f32', 'a'))
 *    ('~f2f16', ('f2f32', 'a@16')) => a
 *    ('pack_half_2x16_split', 'a@32', 'b@32') => ('ior', ('ishl', ('i2i32', ('f2f16', 'b')), 16), ('i2i32', ('f2f16', 'a')))
 */

#ifndef NIR_OPT_ALGEBRAIC_STRUCT_DEFS
#define NIR_OPT_ALGEBRAIC_STRUCT_DEFS

struct transform {
   const nir_search_expression *search;
   const nir_search_value *replace;
   unsigned condition_offset;
};

struct per_op_table {
   const uint16_t *filter;
   unsigned num_filtered_states;
   const uint16_t *table;
};

/* Note: these must match the start states created in
 * TreeAutomaton._build_table()
 */

/* WILDCARD_STATE = 0 is set by zeroing the state array */
static const uint16_t CONST_STATE = 1;

#endif


   static const nir_search_variable search0_0 = {
   { nir_search_value_variable, -1 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3},
};
static const nir_search_expression search0 = {
   { nir_search_value_expression, -1 },
   false,
   -1, 0,
   nir_op_ineg,
   { &search0_0.value },
   NULL,
};

   static const nir_search_constant replace0_0 = {
   { nir_search_value_constant, -1 },
   nir_type_int, { 0x0 /* 0 */ },
};

/* replace0_1 -> search0_0 in the cache */
static const nir_search_expression replace0 = {
   { nir_search_value_expression, -1 },
   false,
   -1, 0,
   nir_op_isub,
   { &replace0_0.value, &search0_0.value },
   NULL,
};

   static const nir_search_variable search1_0 = {
   { nir_search_value_variable, 32 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3},
};

static const nir_search_variable search1_1 = {
   { nir_search_value_variable, 32 },
   1, /* b */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3},
};

static const nir_search_constant search1_2 = {
   { nir_search_value_constant, 32 },
   nir_type_int, { 0x0 /* 0 */ },
};
static const nir_search_expression search1 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_b32csel,
   { &search1_0.value, &search1_1.value, &search1_2.value },
   NULL,
};

   /* replace1_0 -> search1_0 in the cache */

/* replace1_1 -> search1_1 in the cache */
static const nir_search_expression replace1 = {
   { nir_search_value_expression, 32 },
   false,
   0, 1,
   nir_op_iand,
   { &search1_0.value, &search1_1.value },
   NULL,
};

   /* search2_0 -> search1_0 in the cache */

/* search2_1 -> search1_2 in the cache */

/* search2_2 -> search1_1 in the cache */
static const nir_search_expression search2 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_b32csel,
   { &search1_0.value, &search1_2.value, &search1_1.value },
   NULL,
};

   /* replace2_0_0 -> search1_0 in the cache */
static const nir_search_expression replace2_0 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_inot,
   { &search1_0.value },
   NULL,
};

/* replace2_1 -> search1_1 in the cache */
static const nir_search_expression replace2 = {
   { nir_search_value_expression, 32 },
   false,
   0, 1,
   nir_op_iand,
   { &replace2_0.value, &search1_1.value },
   NULL,
};

   /* search3_0 -> search1_0 in the cache */
static const nir_search_expression search3 = {
   { nir_search_value_expression, 8 },
   false,
   -1, 0,
   nir_op_i2i8,
   { &search1_0.value },
   NULL,
};

   /* replace3_0_0 -> search1_0 in the cache */
static const nir_search_expression replace3_0 = {
   { nir_search_value_expression, 16 },
   false,
   -1, 0,
   nir_op_i2i16,
   { &search1_0.value },
   NULL,
};
static const nir_search_expression replace3 = {
   { nir_search_value_expression, 8 },
   false,
   -1, 0,
   nir_op_i2i8,
   { &replace3_0.value },
   NULL,
};

   /* search4_0 -> search1_0 in the cache */
static const nir_search_expression search4 = {
   { nir_search_value_expression, 8 },
   false,
   -1, 0,
   nir_op_u2u8,
   { &search1_0.value },
   NULL,
};

   /* replace4_0_0 -> search1_0 in the cache */
static const nir_search_expression replace4_0 = {
   { nir_search_value_expression, 16 },
   false,
   -1, 0,
   nir_op_u2u16,
   { &search1_0.value },
   NULL,
};
static const nir_search_expression replace4 = {
   { nir_search_value_expression, 8 },
   false,
   -1, 0,
   nir_op_u2u8,
   { &replace4_0.value },
   NULL,
};

   static const nir_search_variable search5_0 = {
   { nir_search_value_variable, 8 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3},
};
static const nir_search_expression search5 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_i2i32,
   { &search5_0.value },
   NULL,
};

   /* replace5_0_0 -> search5_0 in the cache */
static const nir_search_expression replace5_0 = {
   { nir_search_value_expression, 16 },
   false,
   -1, 0,
   nir_op_i2i16,
   { &search5_0.value },
   NULL,
};
static const nir_search_expression replace5 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_i2i32,
   { &replace5_0.value },
   NULL,
};

   /* search6_0 -> search5_0 in the cache */
static const nir_search_expression search6 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_u2u32,
   { &search5_0.value },
   NULL,
};

   /* replace6_0_0 -> search5_0 in the cache */
static const nir_search_expression replace6_0 = {
   { nir_search_value_expression, 16 },
   false,
   -1, 0,
   nir_op_u2u16,
   { &search5_0.value },
   NULL,
};
static const nir_search_expression replace6 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_u2u32,
   { &replace6_0.value },
   NULL,
};

   static const nir_search_variable search7_0 = {
   { nir_search_value_variable, 16 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3},
};
static const nir_search_expression search7 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_f2i32,
   { &search7_0.value },
   NULL,
};

   /* replace7_0_0 -> search7_0 in the cache */
static const nir_search_expression replace7_0 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_f2f32,
   { &search7_0.value },
   NULL,
};
static const nir_search_expression replace7 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_f2i32,
   { &replace7_0.value },
   NULL,
};

   /* search8_0 -> search7_0 in the cache */
static const nir_search_expression search8 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_f2u32,
   { &search7_0.value },
   NULL,
};

   /* replace8_0_0 -> search7_0 in the cache */
/* replace8_0 -> replace7_0 in the cache */
static const nir_search_expression replace8 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_f2u32,
   { &replace7_0.value },
   NULL,
};

   /* search9_0_0 -> search7_0 in the cache */
/* search9_0 -> replace7_0 in the cache */
static const nir_search_expression search9 = {
   { nir_search_value_expression, 16 },
   true,
   -1, 0,
   nir_op_f2f16,
   { &replace7_0.value },
   NULL,
};

   /* replace9 -> search7_0 in the cache */

   /* search10_0 -> search1_0 in the cache */

/* search10_1 -> search1_1 in the cache */
static const nir_search_expression search10 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_pack_half_2x16_split,
   { &search1_0.value, &search1_1.value },
   NULL,
};

   /* replace10_0_0_0_0 -> search1_1 in the cache */
static const nir_search_expression replace10_0_0_0 = {
   { nir_search_value_expression, 16 },
   false,
   -1, 0,
   nir_op_f2f16,
   { &search1_1.value },
   NULL,
};
static const nir_search_expression replace10_0_0 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_i2i32,
   { &replace10_0_0_0.value },
   NULL,
};

static const nir_search_constant replace10_0_1 = {
   { nir_search_value_constant, 32 },
   nir_type_int, { 0x10 /* 16 */ },
};
static const nir_search_expression replace10_0 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_ishl,
   { &replace10_0_0.value, &replace10_0_1.value },
   NULL,
};

/* replace10_1_0_0 -> search1_0 in the cache */
static const nir_search_expression replace10_1_0 = {
   { nir_search_value_expression, 16 },
   false,
   -1, 0,
   nir_op_f2f16,
   { &search1_0.value },
   NULL,
};
static const nir_search_expression replace10_1 = {
   { nir_search_value_expression, 32 },
   false,
   -1, 0,
   nir_op_i2i32,
   { &replace10_1_0.value },
   NULL,
};
static const nir_search_expression replace10 = {
   { nir_search_value_expression, 32 },
   false,
   0, 1,
   nir_op_ior,
   { &replace10_0.value, &replace10_1.value },
   NULL,
};


static const struct transform midgard_nir_lower_algebraic_late_state2_xforms[] = {
  { &search0, &replace0.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state3_xforms[] = {
  { &search1, &replace1.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state4_xforms[] = {
  { &search2, &replace2.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state5_xforms[] = {
  { &search1, &replace1.value, 0 },
  { &search2, &replace2.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state6_xforms[] = {
  { &search3, &replace3.value, 0 },
  { &search5, &replace5.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state7_xforms[] = {
  { &search4, &replace4.value, 0 },
  { &search6, &replace6.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state8_xforms[] = {
  { &search7, &replace7.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state9_xforms[] = {
  { &search8, &replace8.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state11_xforms[] = {
  { &search10, &replace10.value, 0 },
};
static const struct transform midgard_nir_lower_algebraic_late_state12_xforms[] = {
  { &search9, &search7_0.value, 0 },
};

static const struct per_op_table midgard_nir_lower_algebraic_late_table[nir_num_search_ops] = {
   [nir_op_ineg] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         2,
      },
   },
   [nir_op_b32csel] = {
      .filter = (uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (uint16_t []) {
      
         0,
         3,
         4,
         5,
         0,
         3,
         4,
         5,
      },
   },
   [nir_search_op_i2i] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         6,
      },
   },
   [nir_search_op_u2u] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         7,
      },
   },
   [nir_search_op_f2i] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         8,
      },
   },
   [nir_search_op_f2u] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         9,
      },
   },
   [nir_search_op_f2f] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
      },
      
      .num_filtered_states = 2,
      .table = (uint16_t []) {
      
         10,
         12,
      },
   },
   [nir_op_pack_half_2x16_split] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         11,
      },
   },
};

static void
midgard_nir_lower_algebraic_late_pre_block(nir_block *block, uint16_t *states)
{
   nir_foreach_instr(instr, block) {
      switch (instr->type) {
      case nir_instr_type_alu: {
         nir_alu_instr *alu = nir_instr_as_alu(instr);
         nir_op op = alu->op;
         uint16_t search_op = nir_search_op_for_nir_op(op);
         const struct per_op_table *tbl = &midgard_nir_lower_algebraic_late_table[search_op];
         if (tbl->num_filtered_states == 0)
            continue;

         /* Calculate the index into the transition table. Note the index
          * calculated must match the iteration order of Python's
          * itertools.product(), which was used to emit the transition
          * table.
          */
         uint16_t index = 0;
         for (unsigned i = 0; i < nir_op_infos[op].num_inputs; i++) {
            index *= tbl->num_filtered_states;
            index += tbl->filter[states[alu->src[i].src.ssa->index]];
         }
         states[alu->dest.dest.ssa.index] = tbl->table[index];
         break;
      }

      case nir_instr_type_load_const: {
         nir_load_const_instr *load_const = nir_instr_as_load_const(instr);
         states[load_const->def.index] = CONST_STATE;
         break;
      }

      default:
         break;
      }
   }
}

static bool
midgard_nir_lower_algebraic_late_block(nir_builder *build, nir_block *block,
                   const uint16_t *states, const bool *condition_flags)
{
   bool progress = false;

   nir_foreach_instr_reverse_safe(instr, block) {
      if (instr->type != nir_instr_type_alu)
         continue;

      nir_alu_instr *alu = nir_instr_as_alu(instr);
      if (!alu->dest.dest.is_ssa)
         continue;

      switch (states[alu->dest.dest.ssa.index]) {
      case 0:
         break;
      case 1:
         break;
      case 2:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state2_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state2_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 3:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state3_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state3_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 4:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state4_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state4_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 5:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state5_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state5_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 6:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state6_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state6_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 7:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state7_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state7_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 8:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state8_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state8_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 9:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state9_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state9_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 10:
         break;
      case 11:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state11_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state11_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 12:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_lower_algebraic_late_state12_xforms); i++) {
            const struct transform *xform = &midgard_nir_lower_algebraic_late_state12_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      default: assert(0);
      }
   }

   return progress;
}

static bool
midgard_nir_lower_algebraic_late_impl(nir_function_impl *impl, const bool *condition_flags)
{
   bool progress = false;

   nir_builder build;
   nir_builder_init(&build, impl);

   /* Note: it's important here that we're allocating a zeroed array, since
    * state 0 is the default state, which means we don't have to visit
    * anything other than constants and ALU instructions.
    */
   uint16_t *states = calloc(impl->ssa_alloc, sizeof(*states));

   nir_foreach_block(block, impl) {
      midgard_nir_lower_algebraic_late_pre_block(block, states);
   }

   nir_foreach_block_reverse(block, impl) {
      progress |= midgard_nir_lower_algebraic_late_block(&build, block, states, condition_flags);
   }

   free(states);

   if (progress) {
      nir_metadata_preserve(impl, nir_metadata_block_index |
                                  nir_metadata_dominance);
    } else {
#ifndef NDEBUG
      impl->valid_metadata &= ~nir_metadata_not_properly_reset;
#endif
    }

   return progress;
}


bool
midgard_nir_lower_algebraic_late(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[1];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   condition_flags[0] = true;

   nir_foreach_function(function, shader) {
      if (function->impl)
         progress |= midgard_nir_lower_algebraic_late_impl(function->impl, condition_flags);
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 2
 * transforms:
 *    ('fsin', 'a') => ('fsin', ('fdiv', 'a', 3.141592653589793))
 *    ('fcos', 'a') => ('fcos', ('fdiv', 'a', 3.141592653589793))
 */

#ifndef NIR_OPT_ALGEBRAIC_STRUCT_DEFS
#define NIR_OPT_ALGEBRAIC_STRUCT_DEFS

struct transform {
   const nir_search_expression *search;
   const nir_search_value *replace;
   unsigned condition_offset;
};

struct per_op_table {
   const uint16_t *filter;
   unsigned num_filtered_states;
   const uint16_t *table;
};

/* Note: these must match the start states created in
 * TreeAutomaton._build_table()
 */

/* WILDCARD_STATE = 0 is set by zeroing the state array */
static const uint16_t CONST_STATE = 1;

#endif


   static const nir_search_variable search11_0 = {
   { nir_search_value_variable, -1 },
   0, /* a */
   false,
   nir_type_invalid,
   NULL,
   {0, 1, 2, 3},
};
static const nir_search_expression search11 = {
   { nir_search_value_expression, -1 },
   false,
   -1, 0,
   nir_op_fsin,
   { &search11_0.value },
   NULL,
};

   /* replace11_0_0 -> search11_0 in the cache */

static const nir_search_constant replace11_0_1 = {
   { nir_search_value_constant, -1 },
   nir_type_float, { 0x400921fb54442d18L /* 3.141592653589793 */ },
};
static const nir_search_expression replace11_0 = {
   { nir_search_value_expression, -1 },
   false,
   -1, 0,
   nir_op_fdiv,
   { &search11_0.value, &replace11_0_1.value },
   NULL,
};
static const nir_search_expression replace11 = {
   { nir_search_value_expression, -1 },
   false,
   -1, 0,
   nir_op_fsin,
   { &replace11_0.value },
   NULL,
};

   /* search12_0 -> search11_0 in the cache */
static const nir_search_expression search12 = {
   { nir_search_value_expression, -1 },
   false,
   -1, 0,
   nir_op_fcos,
   { &search11_0.value },
   NULL,
};

   /* replace12_0_0 -> search11_0 in the cache */

/* replace12_0_1 -> replace11_0_1 in the cache */
/* replace12_0 -> replace11_0 in the cache */
static const nir_search_expression replace12 = {
   { nir_search_value_expression, -1 },
   false,
   -1, 0,
   nir_op_fcos,
   { &replace11_0.value },
   NULL,
};


static const struct transform midgard_nir_scale_trig_state2_xforms[] = {
  { &search11, &replace11.value, 0 },
};
static const struct transform midgard_nir_scale_trig_state3_xforms[] = {
  { &search12, &replace12.value, 0 },
};

static const struct per_op_table midgard_nir_scale_trig_table[nir_num_search_ops] = {
   [nir_op_fsin] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         2,
      },
   },
   [nir_op_fcos] = {
      .filter = (uint16_t []) {
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 1,
      .table = (uint16_t []) {
      
         3,
      },
   },
};

static void
midgard_nir_scale_trig_pre_block(nir_block *block, uint16_t *states)
{
   nir_foreach_instr(instr, block) {
      switch (instr->type) {
      case nir_instr_type_alu: {
         nir_alu_instr *alu = nir_instr_as_alu(instr);
         nir_op op = alu->op;
         uint16_t search_op = nir_search_op_for_nir_op(op);
         const struct per_op_table *tbl = &midgard_nir_scale_trig_table[search_op];
         if (tbl->num_filtered_states == 0)
            continue;

         /* Calculate the index into the transition table. Note the index
          * calculated must match the iteration order of Python's
          * itertools.product(), which was used to emit the transition
          * table.
          */
         uint16_t index = 0;
         for (unsigned i = 0; i < nir_op_infos[op].num_inputs; i++) {
            index *= tbl->num_filtered_states;
            index += tbl->filter[states[alu->src[i].src.ssa->index]];
         }
         states[alu->dest.dest.ssa.index] = tbl->table[index];
         break;
      }

      case nir_instr_type_load_const: {
         nir_load_const_instr *load_const = nir_instr_as_load_const(instr);
         states[load_const->def.index] = CONST_STATE;
         break;
      }

      default:
         break;
      }
   }
}

static bool
midgard_nir_scale_trig_block(nir_builder *build, nir_block *block,
                   const uint16_t *states, const bool *condition_flags)
{
   bool progress = false;

   nir_foreach_instr_reverse_safe(instr, block) {
      if (instr->type != nir_instr_type_alu)
         continue;

      nir_alu_instr *alu = nir_instr_as_alu(instr);
      if (!alu->dest.dest.is_ssa)
         continue;

      switch (states[alu->dest.dest.ssa.index]) {
      case 0:
         break;
      case 1:
         break;
      case 2:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_scale_trig_state2_xforms); i++) {
            const struct transform *xform = &midgard_nir_scale_trig_state2_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      case 3:
         for (unsigned i = 0; i < ARRAY_SIZE(midgard_nir_scale_trig_state3_xforms); i++) {
            const struct transform *xform = &midgard_nir_scale_trig_state3_xforms[i];
            if (condition_flags[xform->condition_offset] &&
                nir_replace_instr(build, alu, xform->search, xform->replace)) {
               progress = true;
               break;
            }
         }
         break;
      default: assert(0);
      }
   }

   return progress;
}

static bool
midgard_nir_scale_trig_impl(nir_function_impl *impl, const bool *condition_flags)
{
   bool progress = false;

   nir_builder build;
   nir_builder_init(&build, impl);

   /* Note: it's important here that we're allocating a zeroed array, since
    * state 0 is the default state, which means we don't have to visit
    * anything other than constants and ALU instructions.
    */
   uint16_t *states = calloc(impl->ssa_alloc, sizeof(*states));

   nir_foreach_block(block, impl) {
      midgard_nir_scale_trig_pre_block(block, states);
   }

   nir_foreach_block_reverse(block, impl) {
      progress |= midgard_nir_scale_trig_block(&build, block, states, condition_flags);
   }

   free(states);

   if (progress) {
      nir_metadata_preserve(impl, nir_metadata_block_index |
                                  nir_metadata_dominance);
    } else {
#ifndef NDEBUG
      impl->valid_metadata &= ~nir_metadata_not_properly_reset;
#endif
    }

   return progress;
}


bool
midgard_nir_scale_trig(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[1];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   condition_flags[0] = true;

   nir_foreach_function(function, shader) {
      if (function->impl)
         progress |= midgard_nir_scale_trig_impl(function->impl, condition_flags);
   }

   return progress;
}

